/************* script.cpp *************/
#include "script.h"

#include "blind.h"
#include "blind_coding.h"
#include "selective.h"
#include "selective_coding.h"

#include "GaloisField.h"
#include "GaloisFieldElement.h"
#include "GaloisFieldPolynomial.h"
#include "GeneratorPolynomialCreator.h"

using namespace std;
using namespace galois;

vector<int> connected_queue, scaned_queue;

void create_node_xy() {
  for (int i=0; i<nb_nodes_; i++) {
    node_xy *new_node_ = new node_xy();
    new_node_->id_ = i;
    list_xy.push_back(new_node_);
  }
}

void generate_neighbors(int X_MAX, int Y_MAX, int nb_nodes_) {
  node_xy * node_i;
  
  for (int i=0; i<nb_nodes_; i++) {
    node_i = list_xy[i];
    double x = (double(rand())/RAND_MAX)*X_MAX;
    node_i->x = x;
    double y = (double(rand())/RAND_MAX)*Y_MAX;
    node_i->y = y;
    for (int j=0;j<i; j++) {
      node_xy * node_j;
      node_j = list_xy[j];
      double diffx, diffy;
      diffx = (node_i->x-node_j->x)*(node_i->x-node_j->x);
      diffy = (node_i->y-node_j->y)*(node_i->y-node_j->y);
      double distance_ = sqrt(diffx + diffy);
      if (distance_ <= radious_) {
	node_i->list_neighbors_.push_back(j);
	node_j->list_neighbors_.push_back(i);
      }
    }
  }
}

void test_connectivity(node_xy * node_) {
  vector<int>::iterator start, end, int_iter;
  for (int i=0; i<node_->list_neighbors_.size(); i++) {
    int neighbor_ = node_->list_neighbors_[i];
    start = connected_queue.begin();
    end = connected_queue.end();
    int_iter = find(start,end, neighbor_);
    if (int_iter == connected_queue.end()) {
      connected_queue.push_back(neighbor_);
    }
  }
  
  start = scaned_queue.begin();
  end = scaned_queue.end();
  int_iter = find(start,end, node_->id_);
  if (int_iter == scaned_queue.end()) {
    scaned_queue.push_back(node_->id_);
  }
  
  for (int i=0; i<node_->list_neighbors_.size(); i++) {
    int neighbor_ = node_->list_neighbors_[i];
    vector<int>::iterator start = scaned_queue.begin();
    vector<int>::iterator end = scaned_queue.end();
    vector<int>::iterator int_iter;
    int_iter = find(start,end, neighbor_);
    if (int_iter == scaned_queue.end()) {
      test_connectivity(list_xy[neighbor_]);
    }
  }
}

struct option long_options[] = {
  {"help", no_argument, 0, 'h'},
  {"source-placement",                             required_argument, 0, 's'},
  {"nb-sources",                                   required_argument, 0, 'n'},
  {"nb-chunks",                                    required_argument, 0, 'c'},
  {"radious",                                      required_argument, 0, 'r'},
  {"max-x",                                        required_argument, 0, 'x'},
  {"max-y",                                        required_argument, 0, 'y'},
  {"results-directory",                            required_argument, 0, 'd'},
  {"losses",                                       required_argument, 0, 'l'},
  {"run-nb",                                       required_argument, 0, 'b'},
};

int main(int argc, char **argv) {
  while (true) {
    int option_index = 0;
    int c = getopt_long(argc, argv, "hs:n:r:c:x:y:d:l:b:", long_options, &option_index);
    if(c == -1)
      break;
    switch(c) {
    case 'h': 
      cout <<"\n________________________________________________" << endl;
      cout <<"The script can be run as follows:" << endl;
      cout << "./script.sh -c < > -r < > -x < > -y < > n- < > -s < > -l < > -b < >" << endl;
      cout <<"\n Please give the different arguments in the same order as shown in the above example" << endl;

      cout <<"\n________________________________________________" << endl;
      cout << "This script takes 8 inputs in this order:" << endl;
      cout << "c  ------- the number of chunks the file is split to" << endl;
      cout << "r  ------- the transmission radius of nodes" << endl;
      cout << "x  ------- the length of the area where nodes are deployed" << endl;
      cout << "y  ------- the width of the area where nodes are deployed" << endl;
      cout << "s  ------- the placement of the source in the area of deployment: e == extreme & r == random" << endl;
      cout << "n  ------- the number of sources that hold the entire file at time t=0. In this version, we can have from 1 to four sources." << endl;
      cout << "l  ------- the loss rate in the network" << endl;
      cout << "b  ------- the run number. This is useful when we want to run the scripts many times for the same scenario. In this case, new results will not remove previous ones" << endl;
      cout << "d  ------- the directory where the results should be placed. For instance, when c=3, r=2, x=3, y=4, this directory is called \"sim_result_c3_r2_x3_y4\"\n" << endl;
      cout << "do not worry, the directory will be generated automatically by script.sh by providing the values of c,r,x, and y" << endl;

      cout <<"\n________________________________________________" << endl;
      cout << "\n Two important remarks:" << endl;
      cout << "1- The number of nodes is not given by the user. Instead it is computed as a function of r, x, and y" << endl;
      cout << "2- You should read our paper to better understand the assumptions that we make in this simulator" << endl;
      cout << "There are four cooperation strategies that we evaluate in this simulator:" << endl;
      cout << "a- blind              ----> corresponds to blind flooding in our paper" << endl;
      cout << "b- blind_coding       ----> corresponds to blind flooding with network coding" << endl;
      cout << "c- selective          ----> corresponds to selective flooding" << endl;
      cout << "d- selective_coding   ----> corresponds to selective flooding with network coding" << endl;
      exit(-1);
    case 's':
      source_place = *(optarg);
      cout << "the source placement is " << source_place << endl;
      break;
    case 'n':
      nb_sources = atoi(optarg);
      cout << "the number of sources is " << nb_sources << endl;
      break;
    case 'r':
      radious_ = atoi(optarg);
      cout << "the transmission radius is " << radious_ << endl;
      break;
    case 'c':
      nb_chunks_ = atoi(optarg);
      cout << "the number of chunks is " << nb_chunks_ << endl;
      break;
    case 'x':
      X_max_ = atoi(optarg);
      cout << "x_max is " << X_max_ << endl;
      break;
    case 'y':
      Y_max_ = atoi(optarg);
      cout << "y_max is " << Y_max_ << endl;
      break;
    case 'l':
      losses = atoi(optarg);
      cout << "the loss rate is " << losses << endl;
      strcat(b_service_time_,optarg);
      strcat(b_completed_clients_,optarg);
      strcat(b_evolution_chunks_,optarg);
      strcat(b_packet_emitted_,optarg);
      strcat(b_packet_useless_,optarg);
      strcat(b_service_time_,"_run");
      strcat(b_completed_clients_,"_run");
      strcat(b_evolution_chunks_,"_run");
      strcat(b_packet_emitted_,"_run");
      strcat(b_packet_useless_,"_run");
      /**************** Output files for flood_coding ****************/
      strcat(bc_service_time_,optarg);
      strcat(bc_completed_clients_,optarg);
      strcat(bc_evolution_chunks_,optarg);
      strcat(bc_packet_emitted_,optarg);
      strcat(bc_packet_useless_,optarg);
      strcat(bc_service_time_,"_run");
      strcat(bc_completed_clients_,"_run");
      strcat(bc_evolution_chunks_,"_run");
      strcat(bc_packet_emitted_,"_run");
      strcat(bc_packet_useless_,"_run");
      /**************** Output files for bittorrent_coding ****************/
      strcat(sc_service_time_,optarg);
      strcat(sc_service_time_,"_run");
      strcat(sc_completed_clients_,optarg);
      strcat(sc_completed_clients_,"_run");
      strcat(sc_evolution_chunks_,optarg);
      strcat(sc_evolution_chunks_,"_run");
      strcat(sc_packet_emitted_,optarg);
      strcat(sc_packet_emitted_,"_run");
      strcat(sc_packet_useless_,optarg);
      strcat(sc_packet_useless_,"_run");
      /**************** Output files for bittorrent ****************/
      strcat(s_service_time_,optarg);
      strcat(s_service_time_,"_run");
      strcat(s_completed_clients_,optarg);
      strcat(s_completed_clients_,"_run");
      strcat(s_evolution_chunks_,optarg);
      strcat(s_evolution_chunks_,"_run");
      strcat(s_packet_emitted_,optarg);
      strcat(s_packet_emitted_,"_run");
      strcat(s_packet_useless_,optarg);
      strcat(s_packet_useless_,"_run");
      break;
    case 'd':
      directory_results = optarg;
      cout << "the directory result is " << directory_results << endl;
      /**************** Output files for flood ****************/
      strcpy(b_service_time_,directory_results);
      strcat(b_service_time_,"/b_service_time_losses");
      strcpy(b_completed_clients_,directory_results);
      strcat(b_completed_clients_,"/b_completed_clients_losses");
      strcpy(b_evolution_chunks_,directory_results);
      strcat(b_evolution_chunks_,"/b_evolution_chunks_losses");
      strcpy(b_packet_emitted_,directory_results);
      strcat(b_packet_emitted_,"/b_packet_emitted_losses");
      strcpy(b_packet_useless_,directory_results);
      strcat(b_packet_useless_,"/b_packet_useless_losses");
      /**************** Output files for flood_coding ****************/
      strcpy(bc_service_time_,directory_results);
      strcat(bc_service_time_,"/bc_service_time_losses");
      strcpy(bc_completed_clients_,directory_results);
      strcat(bc_completed_clients_,"/bc_completed_clients_losses");
      strcpy(bc_evolution_chunks_,directory_results);
      strcat(bc_evolution_chunks_,"/bc_evolution_chunks_losses");
      strcpy(bc_packet_emitted_,directory_results);
      strcat(bc_packet_emitted_,"/bc_packet_emitted_losses");
      strcpy(bc_packet_useless_,directory_results);
      strcat(bc_packet_useless_,"/bc_packet_useless_losses");
      /**************** Output files for bittorrent_coding ****************/
      strcpy(sc_service_time_,directory_results);
      strcat(sc_service_time_,"/sc_service_time_losses");
      strcpy(sc_completed_clients_,directory_results);
      strcat(sc_completed_clients_,"/sc_completed_clients_losses");
      strcpy(sc_evolution_chunks_,directory_results);
      strcat(sc_evolution_chunks_,"/sc_evolution_chunks_losses");
      strcpy(sc_packet_emitted_,directory_results);
      strcat(sc_packet_emitted_,"/sc_packet_emitted_losses");
      strcpy(sc_packet_useless_,directory_results);
      strcat(sc_packet_useless_,"/sc_packet_useless_losses");
      /**************** Output files for bittorrent ****************/
      strcpy(s_service_time_,directory_results);
      strcat(s_service_time_,"/s_service_time_losses");
      strcpy(s_completed_clients_,directory_results);
      strcat(s_completed_clients_,"/s_completed_clients_losses");
      strcpy(s_evolution_chunks_,directory_results);
      strcat(s_evolution_chunks_,"/s_evolution_chunks_losses");
      strcpy(s_packet_emitted_,directory_results);
      strcat(s_packet_emitted_,"/s_packet_emitted_losses");
      strcpy(s_packet_useless_,directory_results);
      strcat(s_packet_useless_,"/s_packet_useless_losses");
      break;
    case 'b':
      run_nb = atoi(optarg);
      cout << "the run number is " << run_nb << endl;
      /**************** Output files for flood ****************/
      strcat(b_service_time_,optarg);
      strcat(b_completed_clients_,optarg);
      strcat(b_evolution_chunks_,optarg);
      strcat(b_packet_emitted_,optarg);
      strcat(b_packet_useless_,optarg);
      /**************** Output files for flood_coding ****************/
      strcat(bc_service_time_,optarg);
      strcat(bc_completed_clients_,optarg);
      strcat(bc_evolution_chunks_,optarg);
      strcat(bc_packet_emitted_,optarg);
      strcat(bc_packet_useless_,optarg);
      /**************** Output files for bittorrent_coding ****************/
      strcat(sc_service_time_,optarg);
      strcat(sc_completed_clients_,optarg);
      strcat(sc_evolution_chunks_,optarg);
      strcat(sc_packet_emitted_,optarg);
      strcat(sc_packet_useless_,optarg);
      /**************** Output files for bittorrent ****************/
      strcat(s_service_time_,optarg);
      strcat(s_completed_clients_,optarg);
      strcat(s_evolution_chunks_,optarg);
      strcat(s_packet_emitted_,optarg);
      strcat(s_packet_useless_,optarg);
      break;
    default: 
      exit(-1);
    }
  }
  /*********** compute the number of nodes, given the radious, X_max, and Y_max ***********/
  double toto;
  toto = (10*X_max_*Y_max_)/(3.1415*radious_*radious_);
  nb_nodes_ = int(toto);
  if (nb_nodes_ < toto) {nb_nodes_++;}
  cout << "number of nodes is " << nb_nodes_ << endl;
  srand(time(0) % RAND_MAX);

  create_node_xy();
  connected_queue.clear();
  scaned_queue.clear();
  bool first_ = true;
  while (connected_queue.size() < nb_nodes_) {
    if (first_ == true) {
      first_ = false;
    }
    else {
      for (int i=0; i<nb_nodes_; i++) {
	node_xy * node_;
	node_ = list_xy[i];

	node_->list_neighbors_.clear();
      }
    }
    generate_neighbors(X_max_, Y_max_, nb_nodes_);
    
    connected_queue.clear();
    scaned_queue.clear();
    test_connectivity(list_xy[source_id_]);
  }

  /************ run scripts **************/
  if (source_place == 'r') {
    source1 = list_xy[0];
    if (nb_nodes_ >= 2) {
      source2 = list_xy[1];
    }
    if (nb_nodes_ >= 3) {
      source3 = list_xy[2];
    }
    if (nb_nodes_ >= 4) {
      source4 = list_xy[3];
    }
  }
  if (source_place == 'e') {
    cout << "the source is placed at the extremeties" << endl;
    source1 = list_xy[0];
    source1->x = 0; 
    source1->y = 0;
    if (nb_nodes_ >= 2) {
      source2 = list_xy[1];
      source2->x = X_max_; 
      source2->y = Y_max_;
    }
    if (nb_nodes_ >= 3) {
      source3 = list_xy[2];
      source3->x = X_max_; 
      source3->y = 0;
    }
    if (nb_nodes_ >= 4) {
      source4 = list_xy[3];
      source4->x = 0; 
      source4->y = Y_max_;
    }
  }
  
  run_blind();
  run_blind_coding();
  run_selective();
  run_selective_coding();
  
  return 1;
}
